import os
import sys
from typing import AnyStr
from typing import Dict
from typing import Iterable
from typing import TypeVar
from typing import Union

if sys.version_info >= (3, 9):
    PathLike = Union[AnyStr, os.PathLike[AnyStr]]
else:
    PathLike = Union[AnyStr, os.PathLike]

T = TypeVar("T")
Scalar = Union[int, float, bool, str]
IterableOrSingle = Union[Iterable[T], T]

# these types are used to validate schemas internally at type-checking time
SectionT = Dict[str, IterableOrSingle[Scalar]]
InifixConfT = Union[Dict[str, SectionT], SectionT]



import re
from typing import Any
from typing import Mapping

from more_itertools import always_iterable

_PARAM_NAME_REGEXP = re.compile(r"[-\w]+")
SCALAR_TYPES = (int, float, bool, str)


def _uses_invalid_chars(s: str) -> bool:
    ma = re.fullmatch(_PARAM_NAME_REGEXP, s)
    return ma is None


def validate_elementary_item(key: Any, value: Any) -> None:
    if not isinstance(key, str):
        raise ValueError(
            f"Invalid schema: received key '{key}' with type '{type(key)}', "
            "expected a str"
        )
    if len(key) == 0:
        raise ValueError("Invalid schema: received an empty str as key")
    if _uses_invalid_chars(key):
        raise ValueError(
            f"Invalid schema: received key {key!r}, "
            "expected only alphanumeric characters or '-'"
        )
    if not key[0].isalpha():
        raise ValueError(
            f"Invalid schema: received key {key!r}, "
            "keys are expected to start with a letter"
        )
    if not isinstance(value, (*SCALAR_TYPES, list)):
        raise ValueError(
            f"Invalid schema: received value with type '{type(value)}', "
            "exepected an int, float, bool, str, or list"
        )
    for ev in always_iterable(value):
        if not isinstance(ev, SCALAR_TYPES):
            raise ValueError(
                f"Invalid schema: reveived value '{value}' with type '{type(value)}', "
                "exepected a int, float, bool or str"
            )


def validate_inifile_schema(d: Mapping, /) -> None:
    """
    Raise `ValueError` if and only if the argument is not a
    valid configuration.
    """

    for k, v in d.items():
        if not isinstance(k, str):
            raise ValueError(
                f"Invalid schema: received key '{k}' with type '{type(k)}', expected a str"
            )
        if isinstance(v, dict):
            for kk, vv in v.items():
                validate_elementary_item(kk, vv)
        else:
            validate_elementary_item(k, v)



import argparse
import os
import re
import sys
from difflib import unified_diff
from io import StringIO
from tempfile import TemporaryDirectory
from typing import Generator
from typing import List
from typing import Optional
from typing import Sequence
from typing import TextIO
from typing import Union

from inifix.io import _split_tokens
from inifix.io import load

PADDING_SIZE = 2


def _format_section(data: str) -> str:
    lines = data.splitlines()
    contents = []
    comments = []
    parameters = []
    values = []
    for line in lines:
        content, _, comment = line.partition("#")
        content = content.strip()
        if not content and not comment:
            continue
        contents.append(content)
        comments.append(comment.strip())
        if content.startswith("[") or not content:
            continue
        parameter, *value = _split_tokens(content)
        parameters.append(parameter)
        values.append(value)

    column_sizes: List[int] = []
    if not parameters:
        max_name_size = 0
    else:
        max_name_size = max(len(parameter) for parameter in parameters)
        num_value_cols = max(len(_) for _ in values)

        while len(column_sizes) < num_value_cols:
            new_column_idx = len(column_sizes)
            size = 0
            for v in values:
                if len(v) < new_column_idx + 1:
                    continue
                size = max(size, len(v[new_column_idx]))
            column_sizes.append(size + PADDING_SIZE)

    padded_name_col_size = max_name_size + 2 * PADDING_SIZE
    content_size = padded_name_col_size + sum(column_sizes) + PADDING_SIZE

    new_lines = []
    parameter_idx = 0
    for content, comment in zip(contents, comments):
        new_line = ""
        offset = 0
        if content:
            if content.startswith("["):
                new_line += f"{content}\n"
            else:
                new_line += parameters[parameter_idx].ljust(padded_name_col_size)
                for val, size in zip(values[parameter_idx], column_sizes):
                    new_line += val.ljust(size)
                parameter_idx += 1
                new_line = new_line.strip()
                offset = content_size - len(new_line)
        if comment:
            comm = f"# {comment}"
            new_line += comm.rjust(offset + len(comm))
        new_lines.append(new_line)
    res = "\n".join(new_lines)

    return res


def _finalize(res: str) -> str:
    # compress any duplicate new lines
    res = re.sub("\n+", "\n", res)
    # add one empty line before a new section
    res = re.sub(r"(.?)\n\[", r"\1\n\n[", res)
    # ensure there's exactly one newline at the EOF
    res = res.rstrip("\n")
    res += "\n"
    return res


def _iter_sections(fh: Union[StringIO, TextIO]) -> Generator[str, None, None]:
    line = fh.readline()
    while line != "":
        content = [line]
        while (line := fh.readline()) != "" and not line.startswith("["):
            content.append(line)
        yield "".join(content)


def iniformat(fh: Union[StringIO, TextIO, str], /) -> str:
    if isinstance(fh, str):
        fh = StringIO(fh)
    content = []
    for s in _iter_sections(fh):
        content.append(_format_section(s))
    return _finalize("\n".join(content))


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+")
    parser.add_argument(
        "--diff",
        action="store_true",
        help="Print the unified diff to stdout instead of editing files inplace",
    )
    parser.add_argument(
        "--report-noop",
        action="store_true",
        help="Explicitly log noops for files that are already formatted",
    )

    args = parser.parse_args(argv)

    retv = 0

    for file in args.files:
        if not os.path.isfile(file):
            print(f"Error: could not find {file}", file=sys.stderr)
            retv = 1
            continue
        try:
            load(file)
        except ValueError as exc:
            print(f"Error: {exc}", file=sys.stderr)
            retv = 1
            continue

        with open(file) as fh:
            data = fh.read()

        fmted_data = iniformat(data)

        if fmted_data == data:
            if args.report_noop:
                # printing to stderr so that we can pipe into cdiff in --diff mode
                print(f"{file} is already formatted", file=sys.stderr)
            continue
        else:
            retv = 1

        if args.diff:
            for line in unified_diff(
                data.splitlines(), fmted_data.splitlines(), fromfile=file
            ):
                if sys.version_info >= (3, 9):
                    line = line.removesuffix("\n")
                elif line.endswith("\n"):
                    line = line[:-1]
                print(line)
            print("\n")
        else:
            print(f"Fixing {file}", file=sys.stderr)
            if not os.access(file, os.W_OK):
                print(
                    f"Error: could not write to {file} (permission denied)",
                    file=sys.stderr,
                )
                retv = 1
                continue

            try:
                with TemporaryDirectory(dir=os.path.dirname(file)) as tmpdir:
                    tmpfile = os.path.join(tmpdir, "ini")
                    with open(tmpfile, "wb") as bfh:
                        bfh.write(fmted_data.encode("utf-8"))
                    os.replace(tmpfile, file)

            except OSError:
                print(f"Error: could not write to {file}", file=sys.stderr)
                retv = 1
                continue

    return retv


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



import os
import re
from copy import deepcopy
from io import TextIOBase
from tempfile import TemporaryDirectory
from typing import Any
from typing import Callable
from typing import cast
from typing import List
from typing import Literal
from typing import Mapping
from typing import Optional
from typing import TextIO
from typing import Tuple
from typing import Union

from more_itertools import always_iterable
from more_itertools import mark_ends

from inifix._typing import InifixConfT
from inifix._typing import IterableOrSingle
from inifix._typing import PathLike
from inifix._typing import Scalar
from inifix.enotation import ENotationIO
from inifix.validation import SCALAR_TYPES
from inifix.validation import validate_inifile_schema

__all__ = ["load", "loads", "dump", "dumps"]

SECTION_REGEXP = re.compile(r"\[(?P<title>[^(){}\[\]]+)\]\s*")


def bool_caster(s: str) -> bool:
    s = s.lower()
    if s in ("true", "t"):
        return True
    elif s in ("false", "f"):
        return False
    raise ValueError


def str_caster(s: str) -> str:
    if re.match(r"^'.*'$", s) or re.match(r'^".*"$', s):
        return s[1:-1]
    else:
        return s


def _is_numeric(s: str) -> bool:
    try:
        float(s)
    except ValueError:
        return False
    else:
        return True


CASTERS: List[Callable] = [
    int,
    float,
    bool_caster,
    str_caster,
]


class Section(dict):
    def __init__(
        self,
        data: Optional[Mapping[str, IterableOrSingle[Scalar]]] = None,
        /,
        *,
        name: Optional[str] = None,
    ) -> None:
        super().__init__()
        if data is not None:
            for k, v in data.items():
                self[k] = v
        self.name = name

    def __setitem__(self, k: Any, v: Any) -> None:
        if not isinstance(k, str):
            raise TypeError(f"Expected str keys. Received invalid key: {k}")

        _val = deepcopy(v)  # avoid consuming the original iterable
        if not (
            isinstance(v, SCALAR_TYPES)
            or (
                isinstance(_val, list)
                and all(isinstance(_, SCALAR_TYPES) for _ in _val)
            )
        ):
            raise TypeError(
                "Expected all values to be scalars or lists of scalars. "
                f"Received invalid values {v}"
            )

        super().__setitem__(k, v)

    def _dump_to(self, storage: InifixConfT) -> None:
        if self.name is None:
            storage.update(self)
        else:
            storage[self.name] = dict(self)


# load helper functions


def _normalize_data(data: str) -> List[str]:
    # normalize text body `data` to parsable text lines
    lines = []
    for line in data.splitlines():
        # remove comments and normalize whitespace
        line, _, _comment = line.partition("#")
        lines.append(re.sub(r"\s", " ", line.strip()))
    return lines


def _next_token(data: str, pattern: str, start: Literal[0, 1]) -> Tuple[str, int]:
    pos: int = start
    while pos < len(data) and not re.match(pattern, data[pos]):
        pos += 1
    if start == 1:
        end = pos + 1
    else:
        end = pos
    token = data[:end]
    return token, pos


def _split_tokens(data: str) -> List[str]:
    tokens = []
    pattern = r"\s"
    start: Literal[0, 1] = 0
    data = data.strip()
    while True:
        token, pos = _next_token(data, pattern, start)
        tokens.append(token)
        data = data[pos + 1 :].strip()
        if not data:
            break
        if data[0] in ('"', "'"):
            pattern = data[0]
            start = 1
        else:
            pattern = r"\s"
            start = 0
    return tokens


def _tokenize_line(
    line: str, line_number: int, file: Optional[TextIO]
) -> Tuple[str, List[Scalar]]:
    key, *raw_values = _split_tokens(line)
    if not raw_values:
        if file is None:
            raise ValueError(f"Failed to parse line {line_number}: {line!r}")
        else:
            raise ValueError(f"Failed to parse {file}:{line_number}:\n{line}")

    values = []
    for val in raw_values:
        # remove period and trailing zeros to cast to int when possible
        val = re.sub(r"\.0+$", "", val)

        for caster in CASTERS:
            # cast to types from stricter to most permissive
            # `str` will always succeed since it is the input type
            try:
                values.append(caster(val))
                break
            except ValueError:
                continue

    return key, values


def _from_string(data: str, file: Optional[TextIO] = None) -> InifixConfT:
    # see https://github.com/python/mypy/issues/6463
    container: InifixConfT = {}  # type: ignore[assignment]
    lines = _normalize_data(data)
    section = Section()  # the default target is a nameless section
    for line_number, line in enumerate(lines, start=1):
        if not line:
            continue
        match = re.fullmatch(SECTION_REGEXP, line)
        if match is not None:
            if section:
                section._dump_to(container)
            section = Section(name=match["title"])
            continue

        values: Union[Scalar, List[Scalar]]
        key, values = _tokenize_line(line, file=file, line_number=line_number)
        if len(values) == 1:
            values = values[0]
        section[key] = values
    section._dump_to(container)
    return container


def _from_file_descriptor(file: TextIO) -> InifixConfT:
    data = file.read()
    lines = _normalize_data(data)
    if not "".join(lines):
        raise ValueError(f"{file.name!r} appears to be empty.")
    return _from_string(data, file=file)


def _from_path(file: PathLike) -> InifixConfT:
    file = os.fspath(file)
    with open(file) as fh:
        return _from_file_descriptor(fh)


# dump helper functions


def _encode(v: Scalar) -> str:
    if isinstance(v, float):
        return ENotationIO.encode_preferential(v)
    elif isinstance(v, str) and (
        re.search(r"\s", v) is not None
        or v in ("true", "t", "false", "f")
        or _is_numeric(v)
    ):
        return repr(v)
    else:
        return str(v)


def _write_line(key: str, values: IterableOrSingle[Scalar], buffer: TextIO) -> None:
    val_repr = [_encode(v) for v in always_iterable(values)]
    buffer.write(f"{key} {'  '.join([v for v in val_repr])}\n")


def _write_to_buffer(data: InifixConfT, buffer: TextIO) -> None:
    for _is_first, is_last, (key, val) in mark_ends(data.items()):
        if not isinstance(val, dict):
            _write_line(key, val, buffer)
            continue
        buffer.write(f"[{key}]\n")
        for k, v in val.items():
            _write_line(k, v, buffer)
        if not is_last:
            buffer.write("\n")


def _write_to_file(data: InifixConfT, file: PathLike, /) -> None:
    with TemporaryDirectory(dir=os.path.dirname(file)) as tmpdir:
        tmpfile = os.path.join(tmpdir, "ini")
        with open(tmpfile, "w") as fh:
            _write_to_buffer(data, fh)
        os.replace(tmpfile, file)


def load(source: Union[InifixConfT, PathLike, TextIO], /) -> InifixConfT:
    """
    Parse data from a file, or a dict.

    Parameters
    ----------
    source: any of the following
        - a dict (has to be inifix format-compliant)
        - the name of a file to read from, (str, bytes or os.PathLike)
        - a readable file handle (assuming text mode)
    """
    if isinstance(source, TextIOBase):
        source = _from_file_descriptor(source)  # type: ignore [arg-type]
    elif isinstance(source, (str, bytes, os.PathLike)):
        source = _from_path(source)

    source = cast(Mapping, source)

    validate_inifile_schema(source)
    source = cast(InifixConfT, source)
    return source


def loads(source: str, /) -> InifixConfT:
    return _from_string(source)


def dump(data: InifixConfT, /, file: Union[PathLike, TextIOBase]) -> None:
    """
    Write data to a file.

    Parameters
    ----------
    data: dict
        has to be inifix format-compliant

    file: any of the following
        - the name of a file to write to (str, bytes or os.PathLike)
        - a writable file handle (assuming text mode)
    """
    validate_inifile_schema(data)

    try:
        _write_to_buffer(data, file)  # type: ignore
    except AttributeError:
        _write_to_file(data, file)


def dumps(data: InifixConfT, /) -> str:
    from io import StringIO

    s = StringIO()
    dump(data, file=s)
    return s.getvalue()
